#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import argparse
from pprint import pformat
from typing import TYPE_CHECKING

from filetree.service import sync

from modplanner.tty import tty
from modplanner.cli.cmd import modplanner_command
from modplanner.cli import prompt, stage
from modplanner.config.settings import Settings
from modplanner.service import copy, deploy

if TYPE_CHECKING:
    from pathlib import Path

    from filetree import tree


DEPLOY_EXCLUDE_PATTERNS = ["module-index.yaml", ".modulerc"]
DEPLOY_EXCLUDE_DIRS = [".spack"]


def _snapshot_tree(stage_path: Path, snapshot_path: Path):
    tree_builder = copy.copy_modules(
        sources=[stage_path],
        destination=snapshot_path,
        exclude_patterns=DEPLOY_EXCLUDE_PATTERNS,
        exclude_directories=DEPLOY_EXCLUDE_DIRS,
    )
    return tree_builder.build()


def _snapshot_operations(stage_path: Path, snapshot_path: Path):
    snapshot_filetree = _snapshot_tree(
        stage_path=stage_path, snapshot_path=snapshot_path
    )
    io_operations = sync.deploy_filetree(snapshot_filetree, destination=snapshot_path)
    return io_operations


def _link_operations(
    snapshot_source: Path | tree.Tree, deploy_path: Path, settings: Settings
):
    deploy_builder = deploy.deploy_modules(
        sources=[snapshot_source],
        destination=deploy_path,
        layout=settings.layout,
        global_rules=settings.global_rules,
        local_rules=settings.local_rules,
        exclude_patterns=DEPLOY_EXCLUDE_PATTERNS,
        exclude_directories=DEPLOY_EXCLUDE_DIRS,
    )
    deploy_filetree = deploy_builder.build()

    io_operations = sync.deploy_filetree(deploy_filetree, destination=deploy_path)
    return io_operations


def get_deploy_stages(
    settings: Settings, *, dry_run: bool = False
) -> list[stage.Stage]:
    """
    Produce the stages needed for the deploy command,
    namely the stage that copies the merged modules into the
    `snapshot` directory, and the stage that deploys the hirarchical
    structure of categories/profile and linked modules into the `prod` directory.

    The stages hold factory functions to create `OperationBatches` command composites
    at the right moment, which will be executed by the client code.

    The operations are staged lazily because the execution of a stage
    could create some files or directories that a subsequent stage will need,
    and redundant operations are avoided.
    """

    stage_path = settings.paths.stage
    snapshot_path = settings.paths.prod_snapshot
    deploy_path = settings.paths.prod_deploy

    snapshot_stage = stage.Stage(
        name="Snapshot",
        operations_factory=lambda: _snapshot_operations(
            stage_path=stage_path, snapshot_path=snapshot_path
        ),
    )

    if dry_run:
        link_source = _snapshot_tree(stage_path=stage_path, snapshot_path=snapshot_path)
    else:
        link_source = snapshot_path

    link_stage = stage.Stage(
        name="Link",
        operations_factory=lambda: _link_operations(
            snapshot_source=link_source, deploy_path=deploy_path, settings=settings
        ),
    )

    return [snapshot_stage, link_stage]


@modplanner_command("deploy")
class DeployCommand:

    def __init__(self) -> None:
        self._dry_run: bool = False

    def inject_args(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Print the queued operations without executing them.",
        )
        return parser

    def parse_args(self, namespace: argparse.Namespace) -> None:
        if hasattr(namespace, "dry_run"):
            self._dry_run = namespace.dry_run

    @classmethod
    def description(cls) -> str:
        return "Deploy the modulefiles layout to the production target directory."

    @staticmethod
    def _dry_run_exec(settings: Settings) -> None:
        tty.info("Running with dry-run")
        stages = get_deploy_stages(settings=settings, dry_run=True)

        for stage in stages:
            tty.debug(f"Staging operations for stage: {stage.name}")
            stage.stage_operations()
            if not stage:
                tty.success(f"{stage.name}: no operation needed.")
                continue

            prompt.print_operations(stage=stage)

            if not prompt.confirm_stage(stage=stage):
                tty.error("Aborted by the user; exiting.")
                break
            stage.execute()
            tty.success(f"{stage.name}: operations completed.")

    def run(self, settings: Settings) -> None:
        tty.info(f"Running with paths: {pformat(settings.paths.to_dict())}")
        stages = get_deploy_stages(settings=settings, dry_run=self._dry_run)

        for stage in stages:
            tty.debug(f"Staging operations for stage: {stage.name}")
            stage.stage_operations()
            if not stage:
                tty.success(f"{stage.name}: no operation needed.")
                continue

            if self._dry_run:
                prompt.print_operations(stage=stage)
                continue

            if not prompt.confirm_stage(stage=stage):
                tty.error("Aborted by the user; exiting.")
                break

            stage.execute()
            tty.success(f"{stage.name}: operations completed.")
