#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import argparse
from ast import arg
from typing import TYPE_CHECKING

from filetree.service import tree as tree_svc

from modplanner.config.config import ordered_mappings_from_ctx
from modplanner.tty import tty, modules as modules_tty
from modplanner.cli.cmd import modplanner_command, Subcommand
from modplanner.service import parse
from modplanner.service.map import context, engine
from modplanner.service.map.filter import DefaultFilter

if TYPE_CHECKING:
    from modplanner.config.settings import Settings


class QueryDefaults:

    def __init__(self, namespace: argparse.Namespace) -> None:
        if hasattr(namespace, "mapping"):
            self._mapping = namespace.mapping

    def run(self, settings: Settings) -> None:
        default_modules = set()

        class RegisterCallbackFactory(engine.CallbackFactory):

            def __init__(self) -> None:
                super().__init__(None)

            def create_callback(
                self, context: context.ModuleContext
            ) -> engine.ModuleCallback:
                return lambda module: default_modules.add(module)

        for rule in settings.all_rules():
            for factory in rule.filter_factories:
                tmp_filter = factory.create_filter(context.ModuleContext())
                if (
                    isinstance(tmp_filter, DefaultFilter)
                    and tmp_filter._mapping == self._mapping
                ):
                    rule.callback_factories = [RegisterCallbackFactory()]

        stage_path = settings.paths.stage
        stage_tree = tree_svc.parse_tree(stage_path)
        modules = parse.ModuleParser(
            exclude_patterns=[".modulerc", "module-index.yaml"],
            exclude_directories=[".spack"],
        ).modules_from_tree(stage_tree)

        context.build_context(
            modules=modules,
            global_rules=settings.global_rules,
            local_rules=settings.local_rules,
        )

        if default_modules:
            tty.print(f"Modules matching default for {self._mapping}:\n", bold=True)
            modules_tty.print_flat(default_modules)
            return

        tty.err(f"No module matches default for {self._mapping}")


class QueryStructure:

    def run(self, settings: Settings) -> None:
        stage_path = settings.paths.stage
        stage_tree = tree_svc.parse_tree(stage_path)
        modules = parse.ModuleParser(
            exclude_patterns=[".modulerc", "module-index.yaml"],
            exclude_directories=[".spack"],
        ).modules_from_tree(stage_tree)

        module_context = context.build_context(
            modules=modules,
            global_rules=settings.global_rules,
            local_rules=settings.local_rules,
        )
        ordered_mappings = ordered_mappings_from_ctx(
            module_context=module_context, ordering=settings.layout
        )

        modules_tty.print_mappings(
            module_context=module_context, ordered_mappings=ordered_mappings
        )


class QueryPackage:

    def __init__(self, namespace: argparse.Namespace) -> None:
        self._package = namespace.package

    def run(self, settings: Settings) -> None:
        stage_path = settings.paths.stage
        stage_tree = tree_svc.parse_tree(stage_path)
        modules = parse.ModuleParser(
            exclude_patterns=[".modulerc", "module-index.yaml", ".spack"]
        ).modules_from_tree(stage_tree)
        modules = [m for m in modules if m.package_name == self._package]

        if modules:
            module_context = context.build_context(
                modules=modules,
                global_rules=settings.global_rules,
                local_rules=settings.local_rules,
            )
            ordered_mappings = ordered_mappings_from_ctx(
                module_context=module_context, ordering=settings.layout
            )

            modules_tty.print_mappings(
                module_context=module_context, ordered_mappings=ordered_mappings
            )
            return

        tty.err(f"No modules for package: {self._package}")


@modplanner_command("query")
class Query:

    def __init__(self) -> None:
        self._subcommand: Subcommand | None = None

    def inject_args(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        subcommand = parser.add_subparsers(
            dest="query_subcommand",
            required=True,
            description="Subcommands for query command.",
        )

        package = subcommand.add_parser(
            "package", help="Query the mappings of a single package."
        )
        package.add_argument("package", help="Package to query for.")

        defaults = subcommand.add_parser(
            "defaults",
            help="Query the modules that would be mapped as default for a certain mapping.",
        )
        defaults.add_argument("mapping", help="Mapping to query for.")

        subcommand.add_parser(
            "structure", help="Query the entire final filesystem structure."
        )
        return parser

    def parse_args(self, namespace: argparse.Namespace) -> None:
        subcommand = getattr(namespace, "query_subcommand", None)
        if subcommand == "package":
            self._subcommand = QueryPackage(namespace=namespace)
        elif subcommand == "defaults":
            self._subcommand = QueryDefaults(namespace=namespace)
        elif subcommand == "structure":
            self._subcommand = QueryStructure()
        else:
            raise ValueError(f"Unknown subcommand: {subcommand}")

    @classmethod
    def description(cls) -> str:
        return "Query the mapping processor for information about modules."

    def run(self, settings: Settings) -> None:
        assert self._subcommand is not None, "Set subcommand first"
        return self._subcommand.run(settings=settings)
