#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import argparse
from pprint import pformat

from filetree.service import sync

from modplanner.tty import tty
from modplanner.cli.cmd import modplanner_command
from modplanner.cli import prompt, stage
from modplanner.config.settings import Settings
from modplanner.service import copy as source_svc


SOURCE_EXCLUDE_PATTERNS = [".modulerc", "module-index.yaml"]


def get_source_stage(settings: Settings) -> stage.Stage:
    sources_paths = settings.paths.sources
    stage_path = settings.paths.stage

    tty.debug("Merging modules")
    tree_builder = source_svc.copy_modules(
        sources_paths, destination=stage_path, exclude_patterns=SOURCE_EXCLUDE_PATTERNS
    )

    for plg in settings.plugins:
        tty.debug(
            f"Running plugin callbacks for source command: {plg.__class__.__name__}"
        )
        plg.cmd_callbacks("source", builder=tree_builder)

    stage_filetree = tree_builder.build()
    io_operations = sync.deploy_filetree(stage_filetree, destination=stage_path)

    return stage.Stage(
        name="Source", operations_factory=lambda: io_operations, stage=True
    )


@modplanner_command("source")
class SourceCommand:

    def __init__(self) -> None:
        self._dry_run: bool = False

    def inject_args(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Print the queued operations without executing them.",
        )
        return parser

    def parse_args(self, namespace: argparse.Namespace) -> None:
        if hasattr(namespace, "dry_run"):
            self._dry_run = namespace.dry_run

    @classmethod
    def description(cls) -> str:
        return "Source modulefiles to the staging area."

    def run(self, settings: Settings) -> None:
        tty.info(f"Running with paths: \n{pformat(settings.paths.to_dict())}")

        operations_stage = get_source_stage(settings=settings)
        if not operations_stage:
            tty.success("No operation needed.")
            exit(0)

        if self._dry_run:
            prompt.print_operations(operations_stage)
            return

        if not prompt.confirm_stage(operations_stage):
            tty.fatal("Aborted by the user.")

        operations_stage.execute()
        tty.success("Operations completed successfully")
        exit(0)
