#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import dataclasses
import os
from collections import OrderedDict
from pathlib import Path
from typing import TYPE_CHECKING, Protocol, cast

from modplanner.errors import ModplannerError
from modplanner.utils import load_yaml

if TYPE_CHECKING:
    from typing import Any, Iterable
    from filetree import PathLike

    from modplanner.service.map.context import ModuleContext


SCHEMA = Path(os.path.abspath(__file__)).parent / "schemas" / "overloaded_schema.json"


class ConfigValidationError(ModplannerError):
    """
    Error validating config.
    """


def ordered_mappings_from_ctx(
    module_context: ModuleContext, ordering: Iterable[str]
) -> OrderedDict[str, set[str]]:
    mappings = module_context.map_ctx.mappings()
    ret = OrderedDict()
    for ord in ordering:
        if ord in mappings:
            ret[ord] = mappings[ord]
    return ret


class ConfigEntrySpec(Protocol):

    @classmethod
    def from_value(cls, value: Any) -> None: ...
    def update(self, other: ConfigEntrySpec) -> None: ...
    def fill_definitions(self, definitions: DefinitionsSpec) -> None: ...


@dataclasses.dataclass
class PathsSpec:
    root: str | None = dataclasses.field(default=None)
    sources: list[str] | None = dataclasses.field(default=None)
    stage: str | None = dataclasses.field(default=None)
    prod_snapshot: str | None = dataclasses.field(default=None)
    prod_deploy: str | None = dataclasses.field(default=None)

    @classmethod
    def from_value(cls, value: Any) -> PathsSpec:
        value = cast(dict, value)
        return cls(**value)

    def fill_definitions(self, definitions: DefinitionsSpec) -> None:
        for attr in ("root", "sources", "stage", "prod_snapshot", "prod_deploy"):
            if value := getattr(self, attr):
                if _is_definition(value):
                    new_value = definitions[value]
                    setattr(self, attr, new_value)

    def update(self, other: PathsSpec) -> None:
        d = dataclasses.asdict(other)
        for k, v in d.items():
            if v is not None:
                setattr(self, k, v)


def _is_definition(spec: Any) -> bool:
    return isinstance(spec, str) and spec.startswith("def:")


def _fill_rule(rule: dict, definitions: DefinitionsSpec) -> None:
    for m in rule.get("match", []):
        for k, v in m.items():
            if _is_definition(v):
                new_value = definitions[v]
                m[k] = new_value

    for act in rule.get("actions", []):
        for k, v in act.items():
            if _is_definition(v):
                new_value = definitions[v]
                act[k] = new_value


@dataclasses.dataclass
class DefinitionsSpec:
    definitions: dict[str, list | tuple | str | int | float | bool | dict]

    @classmethod
    def from_value(cls, value: Any) -> DefinitionsSpec:
        value = cast(dict, value)
        return cls(definitions=value)

    def fill_definitions(self, definitions: DefinitionsSpec) -> None:
        pass

    def update(self, other: DefinitionsSpec) -> None:
        for definition, value in other.definitions.items():
            self.definitions[definition] = value

    def __getitem__(self, key: str) -> Any:
        # Remove "def:" from key
        name = key[4:]
        try:
            return self.definitions[name]
        except KeyError:
            raise KeyError(f"No definition found for '{name}'.")


@dataclasses.dataclass
class RulesSpec:
    rules: list[dict]

    def __iter__(self):
        yield from self.rules

    @classmethod
    def from_value(cls, value: Any) -> RulesSpec:
        value = cast(list[dict], value)
        return cls(rules=value)

    def fill_definitions(self, definitions: DefinitionsSpec) -> None:
        for rule in self.rules:
            _fill_rule(rule, definitions)

    def update(self, other: RulesSpec) -> None:
        for rule in other.rules:
            if rule not in self.rules:
                self.rules.append(rule)


@dataclasses.dataclass
class ConfigSpec:
    paths: PathsSpec
    layout: list[str]
    definitions: DefinitionsSpec
    global_rules: RulesSpec
    local_rules: dict[str, RulesSpec]
    active_plugins: list[dict]

    def __post_init__(self) -> None:
        self.fill_definitions()

    @classmethod
    def from_value(cls, value: Any) -> ConfigSpec:
        value = cast(dict, value)
        return cls(
            paths=PathsSpec.from_value(value["paths"]),
            layout=value["layout"],
            definitions=DefinitionsSpec.from_value(value["definitions"]),
            global_rules=RulesSpec.from_value(value["global_rules"]),
            local_rules={
                k: RulesSpec.from_value(v) for k, v in value["local_rules"].items()
            },
            active_plugins=value["active_plugins"],
        )

    def fill_definitions(self) -> None:
        self.definitions.fill_definitions(self.definitions)
        self.global_rules.fill_definitions(self.definitions)
        for rule in self.local_rules.values():
            rule.fill_definitions(self.definitions)

    def update(self, other: ConfigSpec) -> None:
        if other.layout:
            self.layout = other.layout
        if other.active_plugins:
            self.active_plugins = list(
                set(self.active_plugins).union(set(other.active_plugins))
            )
        self.paths.update(other.paths)
        self.definitions.update(other.definitions)
        self.global_rules.update(other.global_rules)
        for key, rules in other.local_rules.items():
            if present_rules := self.local_rules.get(key, None):
                present_rules.update(rules)
                continue
            self.local_rules[key] = rules


def merged_config_from_paths(paths: Iterable[PathLike]) -> ConfigSpec:
    dicts = [load_yaml(p) for p in paths]
    specs = [ConfigSpec.from_value(d) for d in dicts]

    if not specs:
        raise ModplannerError("No config found.")

    start = specs[0]
    for other_spec in specs[1:]:
        start.update(other_spec)

    return start
