#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from modplanner.errors import ModplannerError

if TYPE_CHECKING:
    from typing import Callable, Type

    from argparse import ArgumentParser, Namespace

    from filetree.service.tree import FiletreeBuilder

    from modplanner.config.settings import Settings


class PluginsError(ModplannerError):
    """
    Error raised in the plugin module.
    """


class Plugin(Protocol):
    def __init__(self, cfg: dict) -> None: ...
    def inject_args(self, parser: ArgumentParser) -> ArgumentParser: ...
    def parse_args(self, namespace: Namespace) -> None: ...
    def setup(self, settings: Settings) -> None: ...
    def cmd_callbacks(self, cmd: str, builder: FiletreeBuilder) -> None: ...
    def register_processors(self) -> None: ...


_plugins: dict[str, Type[Plugin]] = {}


def plugin(name: str) -> Callable[[Type[Plugin]], Type[Plugin]]:

    def _fn(plg: Type[Plugin]) -> Type[Plugin]:
        _plugins[name] = plg
        return plg

    return _fn


def create(name: str, cfg: dict) -> Plugin:
    if name not in _plugins:
        raise PluginsError(f"Unknown plugin {name}.")
    return _plugins[name](cfg)
