#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol

from modplanner.config import plugin
from modplanner.errors import ModplannerError
from modplanner.service.map import context, engine, filter

if TYPE_CHECKING:
    from modplanner.config.config import ConfigSpec, PathsSpec


@dataclasses.dataclass
class Paths:
    """
    Object holding the paths to operate on.
    """

    root: Path
    # List of paths where to find the modules generated by Spack or BA.
    sources: list[Path]
    # All modules and the module-index.yaml are merged into this directory from the sources.
    stage: Path
    # Path to the "snapshot" of the module deployment, containing the copy of 'stage' at a given moment in time.
    prod_snapshot: Path
    # Path to the hierarchical structure of categories/profile and linked modules from 'prod_repository'.
    prod_deploy: Path

    @classmethod
    def from_spec(cls, spec: PathsSpec) -> Paths:
        assert spec.root is not None
        assert spec.stage is not None
        assert spec.prod_snapshot is not None
        assert spec.prod_deploy is not None

        return cls(
            Path(spec.root).absolute(),
            [Path(s).absolute() for s in spec.sources or []],
            Path(spec.stage).absolute(),
            Path(spec.prod_snapshot).absolute(),
            Path(spec.prod_deploy).absolute(),
        )

    def to_dict(self) -> dict:
        return {
            "root": str(self.root),
            "sources": [str(s) for s in self.sources],
            "stage": str(self.stage),
            "prod_snapshot": str(self.prod_snapshot),
            "prod_deploy": str(self.prod_deploy),
        }


ActionProcessorMap = dict[str, engine.CallbackFactory]
MatchProcessorMap = dict[str, filter.FilterFactory]


class ActionError(ModplannerError):
    """
    Badly formatted action.
    """


class ActionProcessor(Protocol):
    def __call__(self, param: Any) -> engine.CallbackFactory: ...


_action_processors: dict[str, ActionProcessor] = {}


def register_action_processor(name: str, factory: ActionProcessor) -> None:
    _action_processors[name] = factory


def process_action(name: str, param: Any) -> engine.CallbackFactory:
    if name not in _action_processors:
        raise ActionError(f"Unknown callback {name}.")
    return _action_processors[name](param)


class MatchError(ModplannerError):
    """
    Badly formatted match.
    """


class MatchProcessor(Protocol):
    def __call__(self, config: dict) -> filter.FilterFactory: ...


_match_processors: dict[str, MatchProcessor] = {}


def register_match_processor(name: str, func: MatchProcessor) -> None:
    _match_processors[name] = func


def process_match(name: str, config: dict) -> filter.FilterFactory:
    if name not in _match_processors:
        raise MatchError(f"Unknown match type {name}.")
    return _match_processors[name](config)


def _create_rule(rule: dict) -> context.MappingRule:
    exclusive = rule.get("exclusive", False)
    deferred = rule.get("deferred", False)

    filter_factories = []
    for mtch in rule.get("match", []):
        name = mtch.pop("type")
        new_filter_factory = process_match(name, mtch)
        filter_factories.append(new_filter_factory)

    callback_factories = []
    for act in rule.get("actions", []):
        key = next(i for i in act.keys())
        param = act[key]
        new_callback_factory = process_action(key, param)
        callback_factories.append(new_callback_factory)

    return context.MappingRule(
        filter_factories=filter_factories,
        callback_factories=callback_factories,
        exclusive=exclusive,
        deferred=deferred,
    )


class Settings:

    def __init__(
        self,
        paths: Paths,
        layout: list[str],
        global_rules: list[context.MappingRule] | None = None,
        local_rules: dict[str, list[context.MappingRule]] | None = None,
        plugins: list[plugin.Plugin] | None = None,
    ) -> None:
        self.paths = paths
        self.layout = layout
        self.global_rules = global_rules or []
        self.local_rules = local_rules or {}
        self.plugins = plugins or []

    @classmethod
    def from_spec(cls, spec: ConfigSpec) -> Settings:
        paths = Paths.from_spec(spec.paths)
        layout = spec.layout

        plugins = []
        for p in spec.active_plugins:
            for name, cfg in p.items():
                plugins.append(plugin.create(name, cfg))

        instance = cls(paths=paths, layout=layout, plugins=plugins)

        for plg in plugins:
            plg.setup(instance)
            plg.register_processors()

        global_rules = [_create_rule(r) for r in spec.global_rules]
        local_rules = {
            k: [_create_rule(r) for r in v] for k, v in spec.local_rules.items()
        }

        instance.global_rules = global_rules
        instance.local_rules = local_rules

        return instance

    def all_rules(self) -> list[context.MappingRule]:
        return self.global_rules + sum(self.local_rules.values(), start=[])
