#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Main entrypoint for the program.
"""


from __future__ import annotations

import importlib

import argparse

from typing import Protocol

from modplanner.tty import tty
from modplanner.config.config import merged_config_from_paths
from modplanner.config.settings import Settings
from modplanner.cli.cmd import get_commands, get_command
from modplanner.tty.tty import configure_logger

DEFAULT_CONFIG = "./modplanner.yaml"


class ModplannerCommand(Protocol):
    def inject_args(
        self, parser: argparse.ArgumentParser
    ) -> argparse.ArgumentParser: ...
    def parse_args(self, namespace: argparse.Namespace) -> None: ...
    @classmethod
    def description(cls) -> str: ...
    def run(self, settings: Settings) -> None: ...


def _add_main_options(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "-c", "--config", action="append", dest="config", default=[DEFAULT_CONFIG]
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Enable verbose output.", default=0
    )
    return parser


def main():

    try:
        # register all commands and plugins
        importlib.import_module(".cli", "modplanner")
        importlib.import_module(".plugins", "modplanner")

        parser = argparse.ArgumentParser(
            "modplanner",
            description="Map modules into hierarchical filesystem structures "
            "using a YAML declarative interface as input.",
        )
        cmd_subparser = parser.add_subparsers(dest="command", help="Subcommands")
        parser = _add_main_options(parser)

        for command_name, command in get_commands():
            command_parser = cmd_subparser.add_parser(
                command_name, description=command.description()
            )
            command.inject_args(command_parser)

        known_parsed, remaining_args = parser.parse_known_args()

        if known_parsed.verbose == 1:
            configure_logger(level=tty.logging.INFO)
            tty.info("INFO verbosity set")
        elif known_parsed.verbose == 2:
            configure_logger(level=tty.logging.DEBUG)
            tty.debug("DEBUG verbosity set")

        config_paths = known_parsed.config
        config = merged_config_from_paths(config_paths)
        settings = Settings.from_spec(config)

        for plugin in settings.plugins:
            tty.debug(f"Running inject_args for plugin: {plugin.__class__.__name__}")
            parser = plugin.inject_args(parser)

        tty.debug(f"Parsing remaining arguments: {remaining_args}")
        remaining_parsed = parser.parse_args(remaining_args)

        for plugin in settings.plugins:
            tty.debug(f"Running parse_args for plugin: {plugin.__class__.__name__}")
            plugin.parse_args(remaining_parsed)

        command = get_command(known_parsed.command)
        tty.info(f"Running command: {known_parsed.command}")
        command.parse_args(known_parsed)

        command.run(settings=settings)
        tty.print("", space=False)

    except KeyboardInterrupt:
        tty.fatal("Aborted by user; exiting")

    except Exception as e:
        tty.die(e)


if __name__ == "__main__":
    main()
