#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#

"""
Module containing the implementation of the `Module` object.
"""


from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Generator
    from pathlib import Path


@dataclasses.dataclass
class Module:
    # Package name and path are enough info
    # to compute file name and module name.
    package_name: str
    path: Path

    def __hash__(self) -> int:
        return hash(self.module_name)

    @property
    def file_name(self) -> str:
        return self.path.name

    @property
    def module_name(self) -> str:
        return f"{self.package_name}/{self.file_name}"


@dataclasses.dataclass
class Package:
    name: str
    modules: list[Module]


@dataclasses.dataclass
class Packages:
    packages: list[Package]

    def __iter__(self) -> Generator[Package]:
        yield from self.packages

    @classmethod
    def from_modules(cls, modules: Iterable[Module]) -> Packages:
        pack_dict: dict[str, set[Module]] = {}

        for module in modules:
            package_name = module.package_name
            if package_name not in pack_dict:
                pack_dict[package_name] = set()
            pack_dict[package_name].add(module)

        packages = [Package(name, list(mods)) for name, mods in pack_dict.items()]
        return cls(packages)

    def items(self) -> Generator[tuple[str, list[Module]]]:
        return ((p.name, p.modules) for p in self.packages)
