#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Builtins plugin defining standard match processors and action processors.

Basic filters like 'package_name', 'name_pattern' and 'default' are defined here.
Also, basic actions like 'map', 'hide' and 'hide_conditional'.

Other plugins add onto these their own processors.
"""


from __future__ import annotations

from typing import TYPE_CHECKING

from modplanner.tty import tty
from modplanner.service.map.filter import FilterFactory
from modplanner.service.map.engine import CallbackFactory
from modplanner.config.plugin import plugin
from modplanner.config.settings import (
    ActionError,
    register_action_processor,
    register_match_processor,
)
from modplanner.service.map.filter import (
    DefaultFilter,
    NamePatternFilter,
    PackageNameFilter,
)

if TYPE_CHECKING:
    import argparse

    from filetree.service import tree as tree_svc

    from modplanner.config.settings import Settings
    from modplanner.service.map import context, engine, filter


class DefaultFactory(FilterFactory):

    def create_filter(self, context: context.ModuleContext) -> filter.ModuleFilter:
        return DefaultFilter(**self._config, module_context=context)


class NamePatternFactory(FilterFactory):

    def create_filter(self, context: context.ModuleContext) -> filter.ModuleFilter:
        return NamePatternFilter(**self._config)


class PackageNameFactory(FilterFactory):

    def create_filter(self, context: context.ModuleContext) -> filter.ModuleFilter:
        return PackageNameFilter(**self._config)


class MapFactory(CallbackFactory):

    def create_callback(self, context: context.ModuleContext) -> engine.ModuleCallback:
        def _callback(module):
            for key, val in self._param.items():
                if isinstance(val, str):
                    val = [val]
                for v in val:
                    context.map_ctx.map_module(module, key, v)

        return _callback


class HideFactory(CallbackFactory):

    def create_callback(
        self, module_context: context.ModuleContext
    ) -> engine.ModuleCallback:
        if self._param is True:
            return lambda module: module_context.hide_ctx.hide(module)
        return lambda _: None


class ConditionalHideFactory(CallbackFactory):

    def create_callback(
        self,
        module_context: context.ModuleContext,
    ) -> engine.ModuleCallback:

        # For each k: v, substitute v with set(v
        normalize_param = lambda key: {
            k: {v} if isinstance(v, str) else set(v)
            for k, v in self._param.get(key, {}).items()
        }

        mapped: dict[str, set[str]] = normalize_param("mapped")
        not_mapped: dict[str, set[str]] = normalize_param("not_mapped")

        # 'mapped' and 'not_mapped' cannot have common values for the same key
        for m, v in mapped.items():
            if set(v).intersection(set(not_mapped.get(m, []))):
                raise ActionError(
                    "Cannot pass the same mapping to 'mapped' and 'not_mapped' in 'conditional_hide'."
                )

        def _callback(module):

            def _condition(context):
                for k, v in context.items():
                    if mapped and v not in mapped.get(k, []):
                        return False
                    if not_mapped and v in not_mapped.get(k, []):
                        return False
                    return True

            module_context.hide_ctx.add_condition(module, _condition)

        return _callback


@plugin("builtins")
class BuiltinsPlugin:

    def __init__(self, cfg: dict) -> None:
        return

    def inject_args(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        return parser

    def parse_args(self, namespace: argparse.Namespace) -> None:
        pass

    def cmd_callbacks(self, cmd: str, builder: tree_svc.FiletreeBuilder) -> None:
        pass

    def setup(self, settings: Settings) -> None:
        pass

    def register_processors(self) -> None:
        tty.debug("Registering action processors for BultinsPlugin")
        register_action_processor("map", MapFactory)
        register_action_processor("hide", HideFactory)
        register_action_processor("conditional_hide", ConditionalHideFactory)

        tty.debug("Registering match processors for BultinsPlugin")
        register_match_processor("default", DefaultFactory)
        register_match_processor("package_name", PackageNameFactory)
        register_match_processor("name_pattern", NamePatternFactory)
