#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

from typing import TYPE_CHECKING

from modplanner.tty import tty
from modplanner.plugins.spack.spec_repo.db import DBRepo
from modplanner.plugins.spack.index import INDEX_FILE_NAME
from modplanner.plugins.spack.module_repo.index import ModuleIndexRepo

if TYPE_CHECKING:
    from typing import Iterable

    from modplanner.plugins.spack.pool_manager import SpackRuntimeContext
    from modplanner.plugins.spack.spec_repo import SpackSpecRepo
    from modplanner.plugins.spack.module_repo import SpackModuleRepo
    from modplanner import module


class Spack:

    def __init__(self, spec_repo: SpackSpecRepo, module_repo: SpackModuleRepo) -> None:
        self._specs = spec_repo
        self._modules = module_repo

    @classmethod
    def from_context_and_cache(cls, context: SpackRuntimeContext, cache: dict) -> Spack:
        if context.stage_dir in cache:
            tty.debug(
                f"Loading cached module index: {context.stage_dir / INDEX_FILE_NAME}"
            )
            index_repo = cache[context.stage_dir]
        else:
            tty.debug(f"Caching module index: {context.stage_dir / INDEX_FILE_NAME}")
            index_repo = ModuleIndexRepo.from_path(context.stage_dir / INDEX_FILE_NAME)
            cache[context.stage_dir] = index_repo

        db = DBRepo.from_paths(
            install_root=context.stage_dir, spack_root=context.spack_root
        )
        return cls(spec_repo=db, module_repo=index_repo)

    @classmethod
    def from_context(cls, context: SpackRuntimeContext) -> Spack:
        db = DBRepo.from_paths(
            install_root=context.stage_dir, spack_root=context.spack_root
        )
        index_repo = ModuleIndexRepo.from_path(context.stage_dir / INDEX_FILE_NAME)
        return cls(spec_repo=db, module_repo=index_repo)

    @staticmethod
    def _hash_spec(spec) -> str:
        return spec.node_dict_with_hashes()["hash"]

    def _get_modules(self, spec_factory) -> list[module.Module]:
        specs = spec_factory()
        hashes = [self._hash_spec(s) for s in specs]

        modules = []
        for hash_ in hashes:
            if new_module := self._modules.get_module(hash_):
                modules.append(new_module)

        return modules

    def modules_from_constraint(self, constraint: str) -> list[module.Module]:
        return self._get_modules(lambda: self._specs.specs_from_constraint(constraint))

    def modules_from_bulk_constraints(
        self, constraints: Iterable[str]
    ) -> list[module.Module]:
        modules = sum([self.modules_from_constraint(c) for c in constraints], start=[])
        return list(set(modules))

    def implicit_modules(self) -> list[module.Module]:
        return self._get_modules(self._specs.implicit_specs)
