#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TypedDict

from modplanner import module
from modplanner.errors import ModplannerError
from modplanner.utils import BadModuleName, validate_module_name

if TYPE_CHECKING:
    from typing import Generator

Index = dict[str, module.Module]


class IndexEntry(TypedDict):
    path: str
    use_name: str


class BadModuleIndexError(ModplannerError):
    """
    Badly formatted module-index.
    """


INDEX_FILE_NAME = "module-index.yaml"


class ModuleIndex:
    """
    Interface with the information contained in `module-index.yaml` files.

    `__getitem__` excpects a `str` Spack hash, and returns a `Module`.
    """

    def __init__(self, index: Index) -> None:
        self._index = index

    def __contains__(self, hash: str) -> bool:
        return hash in self._index

    def entries(self) -> Generator[tuple[str, module.Module]]:
        yield from self._index.items()

    @classmethod
    def from_dict(cls, dct: dict) -> ModuleIndex:
        sub_dict: dict[str, IndexEntry] = dct.get("module_index", {})
        index = {}
        for hash, module_spec in sub_dict.items():
            module_name = module_spec["use_name"]
            try:
                validate_module_name(module_name)
            except BadModuleName as e:
                raise BadModuleIndexError(f"Module name {e.name} is badly formatted.")
            path = module_spec["path"]
            mod = module.Module(module_name.split("/")[0], Path(path).absolute())
            index[hash] = mod

        return cls(index)

    def to_dict(self) -> dict[str, IndexEntry]:
        ret: dict = {"module_index": {}}
        index: dict[str, IndexEntry] = ret["module_index"]
        for hash, module in self.entries():
            index[hash] = {"path": str(module.path), "use_name": module.module_name}
        return ret

    def get(self, hash: str, default=None) -> module.Module | None:
        return self._index.get(hash, default)
