#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Spack plugin module.

This plugin mostly defines two new filters, 'spack_implicit' and 'spack_specs'.

Also, it provides a callback for the 'source' command, that merges the 'module-index.yaml'
from all module sources into a single yaml file, then written to /stage.
"""


from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from filetree.service import tree as tree_svc
from filetree.fsentry import fsloader

from modplanner.tty import tty
from modplanner.config.plugin import plugin
from modplanner.config.settings import register_match_processor
from modplanner.service.map import filter
from modplanner.plugins.spack import index
from modplanner.plugins.spack.pool_manager import (
    SpackPool,
    SpackConfigContext,
    SpackRuntimeContext,
)

if TYPE_CHECKING:
    import argparse

    from filetree import PathLike

    from modplanner.config.settings import Settings
    from modplanner.service.map import context


class SpackPluginError(Exception):
    """
    Error in Spack plugin.
    """


def _source_indexes(
    context: SpackConfigContext,
    builder: tree_svc.FiletreeBuilder,
    destination: PathLike,
) -> None:
    index_path = Path(context.modules_root / index.INDEX_FILE_NAME)
    if index_path.is_file():
        index_node = tree_svc.build_file(
            index.INDEX_FILE_NAME, loader=fsloader(index_path)
        )
        builder.add_child(index_node, parentpath=destination)


def _source_spack_dbs(
    context: SpackConfigContext,
    builder: tree_svc.FiletreeBuilder,
    destination: PathLike,
) -> None:
    db_path = context.install_root / ".spack-db"
    if not db_path.is_dir():
        raise ValueError(f"DB path {db_path} does not exist")

    if source_node := tree_svc.parse_root(
        context.install_root / ".spack-db", absolute=False
    ):
        builder.add_child(source_node, parentpath=destination)


class ImplicitSpecsFactory(filter.FilterFactory):

    def __init__(self, config: dict, spack_pool: SpackPool) -> None:
        super().__init__(config)
        self._spack_pool = spack_pool

    def create_filter(self, context: context.ModuleContext) -> filter.ModuleFilter:
        modules = self._spack_pool.implicit_modules()
        return filter.ModuleNameFilter(module_names=[m.module_name for m in modules])


class SpackSpecsFactory(filter.FilterFactory):

    def __init__(self, config: dict, spack_pool: SpackPool) -> None:
        super().__init__(config)
        self._spack_pool = spack_pool

    def create_filter(self, context: context.ModuleContext) -> filter.ModuleFilter:
        specs = self._config.get("specs", [])
        modules = self._spack_pool.modules_from_bulk_constraints(specs)
        return filter.ModuleNameFilter(module_names=[m.module_name for m in modules])


@plugin("spack")
class SpackPlugin:

    def __init__(self, cfg: dict) -> None:
        self._module_type = cfg.get("module_type", "tcl")
        self._contexts = [
            SpackConfigContext.from_dict(d) for d in cfg.get("contexts", [])
        ]
        self._settings: Settings | None = None
        self._spack_pool: SpackPool | None = None

    def inject_args(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        return parser

    def parse_args(self, namespace: argparse.Namespace) -> None:
        pass

    def setup(self, settings: Settings) -> None:
        for context in self._contexts:
            if context.modules_root not in settings.paths.sources:
                settings.paths.sources.append(context.modules_root)
        self._settings = settings
        self._spack_pool = SpackPool(
            [
                SpackRuntimeContext.from_config_context_and_settings(
                    ctx, self._settings
                )
                for ctx in self._contexts
            ]
        )

    def register_processors(self) -> None:
        assert self._spack_pool is not None, "Call Plugin.setup"

        tty.debug("Registering match processors for SpackPlugin")
        register_match_processor(
            "spack_implicit",
            lambda cfg: ImplicitSpecsFactory(cfg, spack_pool=self._spack_pool),
        )
        register_match_processor(
            "spack_specs",
            lambda cfg: SpackSpecsFactory(cfg, spack_pool=self._spack_pool),
        )

    def cmd_callbacks(self, cmd: str, builder: tree_svc.FiletreeBuilder) -> None:
        if cmd == "source":
            self._source_callback(builder=builder)

    def _source_callback(self, builder: tree_svc.FiletreeBuilder) -> None:
        assert self._settings is not None, "Run Plugin.setup"

        spack_cache_dir = builder.build_dir(".spack")
        builder.add_child(spack_cache_dir)

        for context in self._contexts:
            identifier = context.str_id()
            destination_node = builder.build_dir(str(identifier))
            builder.add_child(destination_node, spack_cache_dir)
            _source_indexes(
                context=context, builder=builder, destination=destination_node.path
            )
            _source_spack_dbs(
                context=context, builder=builder, destination=destination_node.path
            )
