#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import dataclasses
import hashlib
from pathlib import Path
from multiprocessing import Manager
import concurrent.futures as futures
from typing import TYPE_CHECKING

from modplanner.tty import tty
from modplanner.plugins.spack.api import Spack

if TYPE_CHECKING:
    from typing import Iterable, Any

    from modplanner import module
    from modplanner.config.settings import Settings


@dataclasses.dataclass
class SpackConfigContext:
    version: str
    spack_root: Path
    install_root: Path
    modules_root: Path

    @classmethod
    def from_dict(cls, dct: dict) -> SpackConfigContext:
        try:
            version = dct["version"]
            spack_root = Path(dct["spack_root"]).absolute()
            install_root = Path(dct["install_root"]).absolute()
            modules_root = Path(dct["modules_root"]).absolute()

            return cls(version, spack_root, install_root, modules_root)
        except KeyError:
            raise ValueError(f"Invalid SpackContext dict: {dct}")

    def str_id(self) -> str:
        hashfunc = hashlib.sha256()
        hashfunc.update(self.version.encode("UTF-8"))
        hashfunc.update(str(self.spack_root).encode("UTF-8"))
        hashfunc.update(str(self.install_root).encode("UTF-8"))
        hashfunc.update(str(self.modules_root).encode("UTF-8"))
        return hashfunc.hexdigest()


@dataclasses.dataclass
class SpackRuntimeContext:
    version: str
    spack_root: Path
    stage_dir: Path

    @classmethod
    def from_config_context_and_settings(
        cls, context: SpackConfigContext, settings: Settings
    ) -> SpackRuntimeContext:
        return cls(
            version=context.version,
            spack_root=context.spack_root,
            stage_dir=settings.paths.stage / ".spack" / context.str_id(),
        )


def _modules_from_constraint(
    constraint: str, *, context: SpackRuntimeContext, cache: dict
) -> list[module.Module]:
    spack = Spack.from_context_and_cache(context=context, cache=cache)
    return spack.modules_from_constraint(constraint)


def _modules_from_bulk_constraints(
    constraint: Iterable[str], *, context: SpackRuntimeContext, cache: dict
) -> list[module.Module]:
    spack = Spack.from_context_and_cache(context=context, cache=cache)
    return spack.modules_from_bulk_constraints(constraint)


def _implicit_modules(
    *, context: SpackRuntimeContext, cache: dict
) -> list[module.Module]:
    spack = Spack.from_context_and_cache(context=context, cache=cache)
    return spack.implicit_modules()


class SpackPool:

    def __init__(self, contexts: Iterable[SpackRuntimeContext]):
        self._contexts = list(contexts)
        self._spack_cache = Manager().dict()

    def _pool_exec(self, func, *args, **kwargs) -> Any:
        results: set[Any] = set()
        tty.debug(
            f"Executing parallel call to {func} with {len(self._contexts)} workers"
        )
        with futures.ProcessPoolExecutor(max_workers=len(self._contexts)) as executor:
            fts = [
                executor.submit(
                    func, *args, **kwargs, context=ctx, cache=self._spack_cache
                )
                for ctx in self._contexts
            ]

            for completed in futures.as_completed(fts):
                results = results.union(set(completed.result(timeout=120)))

            return list(results)

    def modules_from_constraint(self, constraint: str) -> list[module.Module]:
        tty.debug(f"Getting modules from constraint: {constraint}")
        return self._pool_exec(_modules_from_constraint, constraint)

    def modules_from_bulk_constraints(
        self, constraints: Iterable[str]
    ) -> list[module.Module]:
        constraints = list(constraints)
        tty.debug(f"Getting modules from constraints: {constraints}")
        return self._pool_exec(_modules_from_bulk_constraints, constraints)

    def implicit_modules(self) -> list[module.Module]:
        tty.debug("Getting implicit modules")
        return self._pool_exec(_implicit_modules)
