#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

from typing import TYPE_CHECKING

from filetree.service import tree as tree_svc

from modplanner.tty import tty
from modplanner.errors import ModplannerError
from modplanner.service import dispatch, parse

if TYPE_CHECKING:
    from typing import Iterable, Optional
    from pathlib import Path

    from .. import module


class ModuleClashError(ModplannerError):
    """
    Module clash detected.
    """


def merge_modules(
    module_pools: Iterable[Iterable[module.Module]],
    *,
    check: bool = False,
    warn: bool = True,
) -> list[module.Module]:
    ret: dict[str, module.Module] = {}
    for pool in module_pools:
        for mod in pool:
            if mod.module_name in ret:
                if check:
                    raise ModuleClashError(
                        f"Module clash detected for '{mod.module_name}'."
                    )
                continue
            ret[mod.module_name] = mod

    return list(ret.values())


def copy_modules(
    sources: Iterable[Path],
    destination: Path,
    *,
    exclude_patterns: Optional[Iterable[str]] = None,
    exclude_directories: Optional[Iterable[str]] = None,
    builder: Optional[tree_svc.FiletreeBuilder] = None,
) -> tree_svc.FiletreeBuilder:
    builder = builder or tree_svc.FiletreeBuilder()

    modules = []
    for src_path in sources:
        tty.debug(f"Parsing modules from: {src_path}")
        parsed_modules = parse.modules_from_path(
            src_path,
            exclude_patterns=exclude_patterns,
            exclude_directories=exclude_directories,
        )
        modules.append(parsed_modules)

    merged_modules = merge_modules(module_pools=modules, check=False)
    builder = dispatch.dispatch_flat_to_path(
        merged_modules, path=destination, builder=builder
    )
    return builder
