#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

from typing import TYPE_CHECKING

from pathlib import Path

from filetree.service import tree as tree_svc
from filetree import PathLike, tree

from modplanner.tty import tty
from modplanner.config.config import ordered_mappings_from_ctx
from modplanner.service import dispatch, parse
from modplanner.service.map.context import MappingRule, build_context

from modplanner.service.copy import merge_modules

if TYPE_CHECKING:
    from typing import Iterable, Optional


def deploy_modules(
    sources: Iterable[PathLike | tree.Tree],
    layout: Iterable[str],
    destination: PathLike,
    global_rules: Iterable[MappingRule],
    local_rules: dict[str, Iterable[MappingRule]],
    *,
    exclude_patterns: Optional[Iterable[str]] = None,
    exclude_directories: Optional[Iterable[str]] = None,
    builder: Optional[tree_svc.FiletreeBuilder] = None,
) -> tree_svc.FiletreeBuilder:
    builder = builder or tree_svc.FiletreeBuilder()

    modules = []
    for src in sources:
        tty.debug(f"Parsing modules from: {src}")
        if isinstance(src, tree.Tree):
            parsed_modules = parse.modules_from_tree(src)
        elif isinstance(src, (str, Path)):
            parsed_modules = parse.modules_from_path(
                src,
                exclude_patterns=exclude_patterns,
                exclude_directories=exclude_directories,
            )
        else:
            raise ValueError(f"Bad source type: {src} ({type(src)})")

        modules.append(parsed_modules)

    merged_modules = merge_modules(modules)

    module_context = build_context(
        global_rules=global_rules, local_rules=local_rules, modules=merged_modules
    )
    ordered_mappings = ordered_mappings_from_ctx(
        module_context=module_context, ordering=layout
    )

    builder = dispatch.dispatch_hierarchy_to_path(
        module_context,
        path=destination,
        ordered_mappings=ordered_mappings,
        strategy=dispatch.LINK,
        builder=builder,
    )

    return builder
