#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Module containing high-level functions to deploy modules on a given
tree structure, getting a `FiletreeBuilder` as a reference and returning it.
"""


from __future__ import annotations

from ast import Call
from typing import TYPE_CHECKING, Protocol

from filetree import fsentry
from filetree.service import tree as tree_svc

from modplanner.tty import tty
from modplanner.module import Module, Packages

if TYPE_CHECKING:
    from collections import OrderedDict
    from typing import Iterable, Optional, Any, Callable

    from filetree import tree, PathLike

    from modplanner.service.map import context


class DispatchStrategy(Protocol):
    """
    Strategy for module dispatch: either link, or copy.
    """

    def __call__(
        self,
        builder: tree_svc.FiletreeBuilder,
        module: Module,
        parent: tree.Node,
    ) -> None: ...


def _copy_strategy(
    builder: tree_svc.FiletreeBuilder, module: Module, parent: tree.Node
) -> None:
    node = builder.build_file(
        name=module.file_name, loader=fsentry.fsloader(module.path)
    )
    builder.add_child(node=node, parent=parent)


def _link_strategy(
    builder: tree_svc.FiletreeBuilder, module: Module, parent: tree.Node
) -> None:
    node = builder.build_link(module.file_name)
    node.entry.target = module.path
    builder.add_child(node=node, parent=parent)


LINK = _link_strategy
COPY = _copy_strategy


def dispatch_modules_flat(
    tree_builder: tree_svc.FiletreeBuilder,
    root: tree.Node[fsentry.Directory],
    modules: Iterable[Module],
    *,
    strategy: DispatchStrategy = COPY,
) -> tree_svc.FiletreeBuilder:
    """
    Flat dispatch of given modules.

    Given the builder, the root node and the modules,
    the produced configuration is the following::

        /root
        │
        ├── package_1
        │   ├── module_1
        │   └── module_2
        │
        └── package_2
            ├── module_3
            └── module_4

    and the modules are either copied or linked, depending on the
    specified strategy.
    """
    modules = set(modules)
    tty.debug(
        f"Computed flat modules dispatch of {len(modules)} modules in {root.path}"
    )
    packages = Packages.from_modules(modules)

    for package_name, modules in packages.items():
        package_dir = tree_builder.build_dir(package_name)
        for module in modules:
            strategy(builder=tree_builder, module=module, parent=package_dir)
        tree_builder.add_child(node=package_dir, parent=root)

    return tree_builder


def dispatch_flat_to_path(
    modules: Iterable[Module],
    path: PathLike,
    *,
    strategy: DispatchStrategy = COPY,
    builder: Optional[tree_svc.FiletreeBuilder] = None,
) -> tree_svc.FiletreeBuilder:
    tree_builder = builder or tree_svc.FiletreeBuilder()
    root = tree_builder.build_dir(str(path))
    tree_builder.set_root(root)

    return dispatch_modules_flat(
        tree_builder=tree_builder,
        root=root,
        modules=modules,
        strategy=strategy,
    )


def _hide_modules(
    builder: tree_svc.FiletreeBuilder,
    parent: tree.Node,
    modules: Iterable[Module],
) -> None:
    modulerc_file = builder.build_file(".modulerc")
    modulerc_file.entry.contents = "\n".join(
        sorted(f"module-hide {m.module_name}" for m in modules)
    )
    builder.add_child(node=modulerc_file, parent=parent)


class TreeBuilderStrategy(Protocol):
    def build_node(self, name: str) -> Any: ...
    def add_child(self, parent: Any, child: Any) -> Any: ...
    def deploy_modules(
        self,
        context: context.ModuleContext,
        parent: Any,
        active_mappings: context.ActiveMappings,
    ) -> None: ...


class FSTreeStrategy:
    def __init__(
        self, tree_builder: tree_svc.FiletreeBuilder, strategy: DispatchStrategy = COPY
    ):
        self.tree_builder = tree_builder
        self.strategy = strategy

    def build_node(self, name: str) -> tree.Node:
        return self.tree_builder.build_dir(name)

    def add_child(self, parent: tree.Node, child: tree.Node) -> tree.Node:
        self.tree_builder.add_child(node=child, parent=parent)
        return child

    def deploy_modules(
        self,
        context: context.ModuleContext,
        parent: tree.Node,
        active_mappings: context.ActiveMappings,
    ) -> None:
        mods = list(context.map_ctx.mapped_modules(active_mappings))
        if hidden := [
            m for m in mods if context.hide_ctx.is_hidden(m, context=active_mappings)
        ]:
            _hide_modules(self.tree_builder, parent, hidden)
        dispatch_modules_flat(self.tree_builder, parent, mods, strategy=self.strategy)


class DictStrategy:
    def __init__(self, callback: Optional[Callable] = None) -> None:
        self._dict = {}
        self._callback = callback

    def get_dict(self) -> dict:
        return self._dict

    def build_node(self, name: str) -> dict:
        return {name: {}}

    def add_child(self, parent: dict, child: dict) -> dict:
        if parent is None:
            parent = self._dict
        key = next(k for k in child)
        val = child.pop(key)
        parent[key] = val
        return val

    def deploy_modules(
        self,
        context: context.ModuleContext,
        parent: dict,
        active_mappings: context.ActiveMappings,
    ):
        assert parent is not None
        mods = list(context.map_ctx.mapped_modules(active_mappings))
        packages = Packages.from_modules(mods)
        for name, modlist in packages.items():
            if self._callback:
                modlist = [self._callback(m, active_mappings) for m in modlist]
            self.add_child(parent=parent, child={name: modlist})


def get_mappings_dict(
    module_context: context.ModuleContext,
    ordered_mappings: OrderedDict[str, set[str]],
    callback: Optional[Callable] = None,
) -> dict:
    strategy = DictStrategy(callback=callback)
    develop_tree(
        ordered_mappings=ordered_mappings,
        module_context=module_context,
        strategy=strategy,
    )
    return strategy.get_dict()


def develop_tree(
    ordered_mappings: OrderedDict[str, set[str]],
    module_context: context.ModuleContext,
    strategy: TreeBuilderStrategy,
) -> None:
    mappings_order = list(ordered_mappings.keys())

    def _walk(
        idx: int = 0,
        parent: Optional[Any] = None,
        active_mappings: Optional[context.ActiveMappings] = None,
    ):
        active_mappings = active_mappings or {}

        if idx >= len(mappings_order):
            strategy.deploy_modules(
                context=module_context,
                parent=parent,
                active_mappings=active_mappings,
            )
            return

        layer_name = mappings_order[idx]
        values = ordered_mappings[layer_name]
        for value in values:
            new_active_mappings = active_mappings.copy()
            new_active_mappings[layer_name] = value
            new_node = strategy.build_node(value)
            child = strategy.add_child(parent=parent, child=new_node)
            _walk(idx + 1, parent=child, active_mappings=new_active_mappings)

    _walk()


def dispatch_hierarchy(
    tree_builder: tree_svc.FiletreeBuilder,
    ordered_mappings: OrderedDict[str, set[str]],
    module_context: context.ModuleContext,
    *,
    strategy: DispatchStrategy = LINK,
) -> tree_svc.FiletreeBuilder:
    """
    Hierarchical dispatch of given modules, based on the given mapping.

    `mappings` **NEEEDS** to be ordered, because the produced directory structure is
    based on this order, and each directory for a given layer contains all values for
    the subsequent mapping.

    E.g., if `mappings` is::

        {
            "foobar": {"foo", "bar"},
            "bazqux": {"baz", "qux"},
        }

    the produced directory structure will be::

        /root
        │
        ├── foo
        │   ├── baz
        │   └── qux
        │
        └── bar
            ├── baz
            └── qux

    and, supposing that we have two modules, mapped::

        module_1: {
            "foobar": {"foo"},
            "bazqux": {"baz"},
        }

        of package "pack1"

    and::

        module_2: {
            "foobar": {"bar"},
            "bazqux": {"baz", "qux"},
        }

        of package "pack2"

    the produced final tree will be::

        /root
        │
        ├── foo
        │   ├── baz
        │   │   └── pack1
        │   └── qux └── module_1
        │
        └── bar
            ├── baz
            │   └── pack2
            └── qux └── module_2
                └── pack2
                    └── module_2

    Again, the strategy will determine whether the modules are linked or copied.
    """
    tty.debug(
        f"Computing hierarchical modules dispatch with layout {ordered_mappings} and context {module_context}"
    )
    build_strategy = FSTreeStrategy(tree_builder=tree_builder, strategy=strategy)

    develop_tree(
        ordered_mappings=ordered_mappings,
        module_context=module_context,
        strategy=build_strategy,
    )

    return tree_builder


def dispatch_hierarchy_to_path(
    module_context: context.ModuleContext,
    path: PathLike,
    ordered_mappings: OrderedDict[str, set[str]],
    *,
    strategy: DispatchStrategy = COPY,
    builder: Optional[tree_svc.FiletreeBuilder] = None,
) -> tree_svc.FiletreeBuilder:
    tree_builder = builder or tree_svc.FiletreeBuilder()
    root = tree_builder.build_dir(str(path))
    tree_builder.set_root(root)

    return dispatch_hierarchy(
        tree_builder=tree_builder,
        module_context=module_context,
        ordered_mappings=ordered_mappings,
        strategy=strategy,
    )
