#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
This module contains the implementations of Context object to keep
track of the modules classifications for the hierarchical deployment
of the production environment.

Concerns are separated into a `MapContext` and a `HideContext`:

- `MapContext` provides an interface to associate modules with `str: set[str]`
mappings, acting like tags tied to a name; e.g. `{"profile": {"chem-phys", "lifesc"}}`.

- `HideContext` provides an interface to hide and show modules, with the possibility
to hide modules based on the mapping environment that is being processed at a given time.
"""


from __future__ import annotations

import dataclasses
from collections import defaultdict
from typing import TYPE_CHECKING, Protocol

from modplanner.service.map.engine import CallbackEngine

if TYPE_CHECKING:
    from typing import Generator, Iterable

    from modplanner import module
    from modplanner.service.map import engine, filter


ModuleMappings = dict[str, set[str]]


class ModuleMapContext:
    """
    Object keeping track of all the mappings for each module.
    The `_mappings` attribute holds a map between a `Module` and
    another map, from `str` to `set[str]`.

    E.g. if the "gromacs" module is mapped as "category: applications"
    and "profile: chem-phys, lifesc", `_mappings` will contain::

        {
            Module("gromacs", path=...): {
                "category": {"applications"},
                "profile": {"chem-phys", "lifesc"}
            },
            ...
        }
    """

    def __init__(self) -> None:
        self._mappings: dict[module.Module, ModuleMappings] = {}

    def mappings(self) -> dict[str, set[str]]:
        """
        Return a map of all mappings -> values appearing in the
        `_mappings` attribute.
        """
        ret: dict[str, set[str]] = defaultdict(set)
        for mappings in self._mappings.values():
            for mapping, values in mappings.items():
                ret[mapping] = ret[mapping].union(values)
        return ret

    def all_mappings(self) -> dict[module.Module, ModuleMappings]:
        return self._mappings

    def map_module(self, module: module.Module, mapping: str, value: str) -> None:
        """
        Register a mapping for a module.
        """
        if module not in self._mappings:
            self._mappings[module] = {}
        if mapping not in self._mappings[module]:
            self._mappings[module][mapping] = set()
        self._mappings[module][mapping].add(value)

    def get_mappings(self, module: module.Module) -> ModuleMappings:
        """
        Get all mappings for a module.
        """
        return self._mappings.get(module, {})

    def get_mappings_from_name(self, module_name: str) -> ModuleMappings:
        """
        Get all mappings for a module from a module_name.
        """
        for module, mappings in self._mappings.items():
            if module.module_name == module_name:
                return mappings
        return {}

    def get_mappings_from_package_name(
        self, package_name: str
    ) -> dict[module.Module, ModuleMappings]:
        """
        Get all mappings for multiple modules from a package_name.
        """
        ret = {}
        for module, mappings in self._mappings.items():
            if module.package_name == package_name:
                ret[module] = mappings
        return ret

    def is_mapped(self, module: module.Module, maps: ActiveMappings) -> bool:
        """
        Test whether a module is mapped in a certain way.

        `MapCtx` is a `str: str` map, and the entire context needs to be satisfied.

        E.g. if `MapCtx` is `{"foo": "bar", "baz": "qux"}`, the method returns `True`
        only if **BOTH** mappings exist for the given module.
        """
        if not maps:
            return False

        if module not in self._mappings:
            return False

        mappings = self._mappings[module]
        if any((v not in mappings.get(m, set()) for m, v in maps.items())):
            return False

        return True

    def has_mapping(self, module: module.Module, mapping: str) -> bool:
        """
        Test whether a module has a mapping.

        For this method to return `True`, it is enough that the
        given module has **ANY** mapping for the name passed as argument.
        """
        if module not in self._mappings:
            return False
        return mapping in self._mappings[module]

    def mapped_modules(self, mappings: ActiveMappings) -> Generator[module.Module]:
        """
        Yields a generator of modules mapped as the given context
        (a `Generator` is a one-shot lazy iterator).

        As above, the entire context needs to be satisfied for the
        module to be included in the result.
        """
        for module, maps in self._mappings.items():
            for k, v in mappings.items():
                if v not in maps.get(k, set()):
                    break
            # Python quirk: else runs only if the loop is not broken.
            else:
                yield module


ActiveMappings = dict[str, str]


class HideCondition(Protocol):
    """
    Interface for the hide condition functions.
    Given some active mappings (where we are now, so to speak), return a `bool`.
    """

    def __call__(self, active_mappings: ActiveMappings) -> bool: ...


class ModuleHideContext:
    """
    Object keeping track of whether a module is hidden or not.

    A global `_hidden` attribute keeps track of `Module`s that are hidden everywhere.
    A `_hide_conditions` map maps each `Module` to a `list` of functions, that return,
    when called with the current mapping context as parameter, whether the module is
    hidden in that particular context.

    E.g. if we are currently processing the modules mapped "foo": "bar" and "bar": "baz",
    each hide condition mapped to each module will be called with::

        {"foo": "bar", "bar": "baz"}

    as argument, and will return `True` if the module is hidden in the::

        {"foo": "bar", "bar": "baz"}

    context.
    """

    def __init__(self) -> None:
        self._always_hidden: set[module.Module] = set()
        self._hide_conditions: dict[module.Module, list[HideCondition]] = {}

    def hide(self, module: module.Module) -> None:
        """
        Hide a module everywhere.
        """
        self._always_hidden.add(module)

    def add_condition(self, module: module.Module, condition: HideCondition) -> None:
        """
        Add a single `HideCondition` for a given module.
        """
        if module not in self._hide_conditions:
            self._hide_conditions[module] = []

        self._hide_conditions[module].append(condition)

    def show(self, module: module.Module, *, clear_conditions: bool = False) -> None:
        """
        Show a single module, if it was hidden before.

        If `clear_conditions` is `True`, also clear all `HideConditions` for the given module.
        """
        if module in self._always_hidden:
            self._always_hidden.remove(module)

        if clear_conditions and module in self._hide_conditions:
            del self._hide_conditions[module]

    def is_hidden(
        self, module: module.Module, *, context: ActiveMappings | None = None
    ) -> bool:
        """
        Returns whether the module is hidden.

        If a context is provided, also check all `hide_conditions` for the given module.
        """
        if module in self._always_hidden:
            return True

        if context is not None and (conds := self._hide_conditions.get(module, [])):
            if any(c(context) for c in conds):
                return True

        return False


class ModuleContext:
    """
    Just a wrapper around `ModuleMapContext` and `ModuleHideContext`.
    """

    def __init__(self) -> None:
        self.map_ctx = ModuleMapContext()
        self.hide_ctx = ModuleHideContext()


@dataclasses.dataclass
class MappingRule:
    """
    Simple container for a `CallbackEngine` spec.

    Filter and callback factories create `Filter` and `ModuleCallback`
    objects when called with a specific `ModuleContext` as the only argument.
    """

    filter_factories: list[filter.FilterFactory]
    callback_factories: list[engine.CallbackFactory]
    exclusive: bool
    deferred: bool


class ModuleContextBuilder:
    """
    Builder pattern for a `ModuleContext`.

    The idea is that the global rules are processed first, and modifications to the
    module pool executed by global rules are propagated to all following rules.

    See the `build_context` method for more information.
    """

    def __init__(self) -> None:
        self._context = ModuleContext()

    def _build_engine(
        self, rule: MappingRule, module_pool: set[module.Module]
    ) -> engine.ICallbackEngine:
        """
        Build a single `CallbackEngine` from a given `MappingRule` specification.
        """
        engine = CallbackEngine(
            factory.create_filter(self._context) for factory in rule.filter_factories
        )
        for factory in rule.callback_factories:
            callback = factory.create_callback(self._context)
            engine.register_callback(callback)

        if rule.exclusive:
            # We pass the module pool by reference and, if the rule is exclusive,
            # we add a callback to remove the module from the pool when it is matched.

            # FIXME: I don't like the fact that we modify objects passed by reference
            # to multiple functions, but I don't know if there is a better way.
            remove_callback = lambda module: module_pool.remove(module)
            engine.register_callback(remove_callback)

        return engine

    def _apply_rules(
        self, rules: Iterable[MappingRule], module_pool: set[module.Module]
    ) -> None:

        def _key(rule: MappingRule):
            if rule.exclusive:
                return (0, rule.deferred)
            elif not rule.exclusive and not rule.deferred:
                return (1, 0)
            else:
                return (2, 0)

        rules = sorted(rules, key=_key)

        # Sort order:
        #   - exclusive, non-deferred rules
        #   - exclusive, deferred rules
        #   - non-exclusive, non-deferred rules
        #   - deferred rules

        # For each rule, build the engine and run it on the provided module pool.
        # (this will of course manipulate the pool via callbacks if necessary,
        # as described above.)

        for rule in rules:
            callback_engine = self._build_engine(rule=rule, module_pool=module_pool)
            callback_engine.apply(module_pool)

    def build_context(
        self,
        modules: Iterable[module.Module],
        global_rules: Iterable[MappingRule],
        local_rules: dict[str, Iterable[MappingRule]],
    ) -> ModuleContext:
        # Iterable[] can be a Generator (a one-shot iterable iterator),
        # so we need to persist the references in memory.
        starting_pool = set(modules)

        # Apply global rules first, acting on the "starting" pool,
        # which will be passed below to the local rules after being
        # processed.

        # If a global rule removes modules from the pool because
        # we don't want to processes them further, it happens here.
        self._apply_rules(rules=global_rules, module_pool=starting_pool)

        for rule_batch in local_rules.values():
            # For each batch of local rules, we create a new "local" pool
            # starting from the "main" pool, so that every manipulation
            # of the given pool is restricted to the rule batch.

            # E.g. if we are processing a rule batch for the "category" mapping,
            # exclusive rules will remove modules from the pool until the batch
            # has been processed. Afterwards, the "profile" batch will start from
            # the "main" pool again.
            local_pool = set(starting_pool)
            self._apply_rules(rules=rule_batch, module_pool=local_pool)

        # Reset the attributes and return the built ModuleContext.
        ret = self._context
        self._context = ModuleContext()
        return ret


_default_builder = ModuleContextBuilder()
build_context = _default_builder.build_context
