#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Module containing the interface and implementations of the `CallbackEngine`
and `EngineCallback` objects.

`EngineCallback` objects take a `Module` as argument and return None.

`CallbackEngine` objects hold a list of `ModuleFilter` and `EngineCallback` objects,
and apply all callbacks on a given module if all the filters of the given engine
return `True` when called on the module.
"""


from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing import Iterable, Any

    from modplanner import module
    from modplanner.service.map import context, filter


class ModuleCallback(Protocol):
    """
    General interface for an engine callback:
    simple callable that takes a module and returns None.
    """

    def __call__(self, module: module.Module) -> None: ...


class CallbackFactory(abc.ABC):

    def __init__(self, param: Any) -> None:
        self._param = param

    @abc.abstractmethod
    def create_callback(self, context: context.ModuleContext) -> ModuleCallback: ...


class ICallbackEngine(Protocol):
    """
    Interface for the callback engine; this exists only for testing purposes.
    """

    def apply(self, modules: Iterable[module.Module]) -> None: ...


class CallbackEngine:
    """
    Simple engine for filters and callbacks.

    Callbacks are selectively applied in the `apply` method to
    all `Module` objects that match all the filters.
    """

    def __init__(self, filters: Iterable[filter.ModuleFilter]) -> None:
        self._filters = list(filters)
        self._callbacks: list[ModuleCallback] = []

    def register_callback(self, callback: ModuleCallback) -> None:
        self._callbacks.append(callback)

    def apply(self, modules: Iterable[module.Module]) -> None:
        """
        Loop over the provided modules, when::

            filter(module) is True

        for all filters in `self._filters`, all callbacks in `self._callbacks`
        are called on the given module.
        """
        module_pool = set(modules)
        for module in module_pool:
            if all(flt(module) for flt in self._filters):
                for callback in self._callbacks:
                    callback(module)
