#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Module containing the interface and implementations for `ModuleFilter` objects.

These are objects that, when called with a `Module` as argument, return a bool.
"""


from __future__ import annotations

import re
import abc
from typing import TYPE_CHECKING, Protocol

from modplanner.errors import ModplannerError

if TYPE_CHECKING:
    from typing import Iterable

    from modplanner.module import Module
    from modplanner.service.map import context


class FilterError(ModplannerError):
    """
    Error raised from the filter module.
    """


class ModuleFilter(Protocol):
    """
    Interface for a filter that takes a `Module` as input and returns a bool.
    """

    def __call__(self, module: Module) -> bool: ...


class PackageNameFilter:
    """
    Implementation of a filter that takes a `Module` as input and returns
    whether the package name of the module matches any of the given names.
    """

    def __init__(self, package_names: Iterable[str]) -> None:
        self._names = package_names

    def __str__(self) -> str:
        return f"PackageNameFilter({self._names})"

    def __call__(self, module: Module) -> bool:
        return module.package_name in self._names


class NamePatternFilter:
    """
    Implementation of a filter that takes a `Module` as intput and returns
    whether the module name matches any of the given patterns.
    """

    def __init__(self, name_patterns: Iterable[str]) -> None:
        self._regex_engines = [re.compile(p) for p in name_patterns]
        self._patterns = name_patterns

    def __str__(self) -> str:
        return f"NameRegexFilter({self._patterns})"

    def __call__(self, module: Module) -> bool:
        return any(bool(e.search(module.module_name)) for e in self._regex_engines)


class DefaultFilter:
    """
    Implementation of a filter that takes a `Module` as intput and returns
    whether the module is already mapped with the given mapping.
    """

    def __init__(self, mapping: str, module_context: context.ModuleContext) -> None:
        self._mapping = mapping
        self._context = module_context

    def __call__(self, module: Module) -> bool:
        return not self._context.map_ctx.has_mapping(module, self._mapping)


class ModuleEqFilter:
    """
    Implementation of a filter that takes a `Module` as intput and returns
    whether the module appears in a list of stored modules.
    """

    def __init__(self, modules: Iterable[Module]) -> None:
        self._modules = modules

    def __call__(self, module: Module) -> bool:
        return module in self._modules


class ModuleNameFilter:
    """
    Implementation of a filter that takes a `Module` as intput and returns
    whether its name appears in a list of stored names.
    """

    def __init__(self, module_names: Iterable[str]) -> None:
        self._names = list(module_names)

    def __call__(self, module: Module) -> bool:
        return module.module_name in self._names


class FilterFactory(abc.ABC):

    def __init__(self, config: dict) -> None:
        self._config = config

    @abc.abstractmethod
    def create_filter(self, context: context.ModuleContext) -> ModuleFilter: ...
