#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
This module contains the module parsing logic.

Given a `Tree` object, the `ModuleParser` returns a list of `Module` objects.
"""


from __future__ import annotations

import re
from typing import TYPE_CHECKING, Protocol

from filetree import fsentry
from filetree.service import tree as tree_svc

from modplanner.tty import tty
from modplanner.module import Module

if TYPE_CHECKING:
    from typing import Iterable

    from filetree import tree, PathLike


class ModuleParserFunc(Protocol):
    """
    A function to parse a `tree.Node` into a `Module`.
    This interface exists mainly for testing purposes.
    """

    def __call__(self, node: tree.Node) -> Module: ...


def flat_parser(node: tree.Node) -> Module | None:
    """
    A simple flat parser that returns a `Module` if the node is a leaf,
    and its parent is a directory, and the node itself is not a directory.
    """
    if (
        node.is_leaf()
        and node.parent is not None
        and not isinstance(node.entry, fsentry.Directory)
    ):
        package_name = node.parent.name
        path = node.path
        return Module(package_name, path)
    return None


class ModuleParser:
    """
    Simple module parser, `parse_tree` takes a `Tree`
    and returns a `Module` `Generator`, calling `parser`
    for each `Node` in the `Tree`.
    """

    def __init__(
        self,
        parser: ModuleParserFunc | None = None,
        exclude_patterns: Iterable[str] | None = None,
        exclude_directories: Iterable[str] | None = None,
    ) -> None:
        self.exclude_patterns = (
            [re.compile(p) for p in exclude_patterns]
            if exclude_patterns is not None
            else []
        )
        self.exclude_directories = (
            list(exclude_directories) if exclude_directories is not None else []
        )
        self.parser = parser or flat_parser

    def modules_from_tree(self, tree_: tree.Tree) -> list[Module]:
        ret = []
        excluded_pattern = lambda node: any(
            p.search(node.name) for p in self.exclude_patterns
        )
        excluded_dir = lambda node: any(
            node.name == d for d in self.exclude_directories
        )
        for node in tree_.breadth_first_nodes(stop_condition=excluded_dir):
            if not excluded_pattern(node) and (module := self.parser(node)):
                ret.append(module)

        if ret and tree_.root:
            tty.debug(f"Parsed {len(ret)} modules from {tree_.root.path}")

        return ret


def modules_from_path(
    path: PathLike,
    *,
    exclude_patterns: Iterable[str] | None = None,
    exclude_directories: Iterable[str] | None = None,
) -> list[Module]:
    parser = ModuleParser(
        exclude_directories=exclude_directories, exclude_patterns=exclude_patterns
    )
    tree = tree_svc.parse_tree(path)
    return parser.modules_from_tree(tree)


def modules_from_tree(
    tree: tree.Tree,
    *,
    exclude_patterns: Iterable[str] | None = None,
    exclude_directories: Iterable[str] | None = None,
) -> list[Module]:
    parser = ModuleParser(
        exclude_directories=exclude_directories, exclude_patterns=exclude_patterns
    )
    return parser.modules_from_tree(tree)
