#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import os
import dataclasses

from typing import TYPE_CHECKING

from rich.table import Table
from rich.text import Text
from filetree.operations import (
    ExecutableBatch,
    RmDir,
    RmFile,
    RmLink,
    MkDir,
    WriteFile,
    Link,
    Chmod,
)

from modplanner.tty import tty

if TYPE_CHECKING:
    from typing import Iterable


REMOVE_COLOR = "red1"
MAKE_COLOR = "gold3"
LINK_COLOR = "turquoise2"
CHMOD_COLOR = "light_steel_blue"


@dataclasses.dataclass
class ClassifiedOperations:
    remove_dirs: list[RmDir]
    remove_files: list[RmFile]
    remove_links: list[RmLink]
    create_dirs: list[MkDir]
    write_files: list[WriteFile]
    create_links: list[Link]
    change_permissions: list[Chmod]

    @classmethod
    def from_operations(cls, operations: ExecutableBatch) -> ClassifiedOperations:
        rm_d = []
        rm_f = []
        rm_l = []
        mk_d = []
        wr_f = []
        cr_l = []
        ch_p = []

        for op in operations:
            if isinstance(op, RmDir):
                rm_d.append(op)
            elif isinstance(op, RmFile):
                rm_f.append(op)
            elif isinstance(op, RmLink):
                rm_l.append(op)
            elif isinstance(op, MkDir):
                mk_d.append(op)
            elif isinstance(op, WriteFile):
                wr_f.append(op)
            elif isinstance(op, Link):
                cr_l.append(op)
            elif isinstance(op, Chmod):
                ch_p.append(op)

        return ClassifiedOperations(
            remove_dirs=rm_d,
            remove_files=rm_f,
            remove_links=rm_l,
            create_dirs=mk_d,
            write_files=wr_f,
            create_links=cr_l,
            change_permissions=ch_p,
        )


def _print_paths(title: str, paths: Iterable[str], color: str) -> None:
    table = Table("title", "one", show_header=False, box=None)
    for path in paths:
        table.add_row(Text(title + ":", style=color), Text(path, style=color))
    tty.print_renderable(table, indentation=4)


def _print_with_arrows(
    title: str, paths: Iterable[tuple[str, str]], color: str, arrow_char: str
) -> None:
    table = Table("title", "one", "arrow", "two", show_header=False, box=None)
    for left, right in paths:
        table.add_row(
            Text(title + ":", style=color),
            Text(left, style=color),
            Text(arrow_char, style=color),
            Text(right, style=color),
        )
    tty.print_renderable(table, indentation=4)


def print_operations(operations: ExecutableBatch) -> None:
    classified_ops = ClassifiedOperations.from_operations(operations)
    cwd = os.getcwd()

    pairings = [
        ("remove_dirs", "RM Dir", REMOVE_COLOR),
        ("remove_files", "RM File", REMOVE_COLOR),
        ("remove_links", "RM Link", REMOVE_COLOR),
        ("create_dirs", "MK Dir", MAKE_COLOR),
        ("write_files", "WRITE File", MAKE_COLOR),
    ]

    for attr, text, color in pairings:
        if ops := getattr(classified_ops, attr):
            _print_paths(
                text,
                [str(op.node.path.relative_to(cwd)) for op in ops],
                color=color,
            )

    if ops := classified_ops.create_links:
        paths = [
            tuple(
                (str(op.node.path.relative_to(cwd)), str(op.target().relative_to(cwd)))
            )
            for op in ops
        ]
        _print_with_arrows("Link", paths, color=LINK_COLOR, arrow_char="<-")

    if ops := classified_ops.change_permissions:
        paths = [
            tuple((str(op.node.path.relative_to(cwd)), str(op.mode))) for op in ops
        ]
        _print_with_arrows("Chmod", paths, color=CHMOD_COLOR, arrow_char="->")
