#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Optional

import logging
from rich.console import Console, RenderableType
from rich.padding import Padding
from rich.prompt import Prompt
from rich.style import Style
from rich.text import Text

if TYPE_CHECKING:
    from typing import Optional


STYLE_KEY = "__style__"


@dataclasses.dataclass
class TextStyle:
    indentation: int = 0
    color: str | None = None
    background_color: str | None = None
    bold: bool = False
    italic: bool = False
    underline: bool = False
    strikethrough: bool = False

    def copy(self, **kwargs) -> TextStyle:
        params = dataclasses.asdict(self)
        params.update(kwargs)
        return TextStyle(**params)


def stylize(
    msg: str,
    text_style: TextStyle,
) -> Text:
    text = Text(msg)
    if text_style:
        if text_style.indentation:
            text.pad_left(text_style.indentation)
        style = Style(
            color=text_style.color,
            bgcolor=text_style.background_color,
            bold=text_style.bold,
            underline=text_style.underline,
            italic=text_style.italic,
            strike=text_style.strikethrough,
        )
        text.stylize(style)
    return text


class RichTty:

    def __init__(
        self,
        style: TextStyle | None = None,
        console: Console | None = None,
        *,
        linespaces: int = 1,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        self._linespaces = linespaces
        self._console = console or Console()
        self._style = style or TextStyle()
        self._prompt = Prompt()
        self._logger = logger or logging.getLogger(__name__)
        self._lastchar = None

    def debug(self, msg, *args, **kwargs) -> None:
        return self._logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs) -> None:
        return self._logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs) -> None:
        return self._logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs) -> None:
        return self._logger.error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs) -> None:
        return self._logger.critical(msg, *args, **kwargs)

    def padding(self, msg: RenderableType, style: TextStyle) -> Padding:
        # TODO: fix
        return Padding(msg, (0, style.indentation))

    def print(
        self,
        msg: str,
        *,
        indentation: int = 0,
        color: str | None = None,
        background_color: str | None = None,
        bold: bool = False,
        italic: bool = False,
        underline: bool = False,
        strikethrough: bool = False,
        space: bool = True,
    ) -> None:
        if any(
            (
                indentation,
                color,
                background_color,
                bold,
                italic,
                underline,
                strikethrough,
            )
        ):
            sty = TextStyle(
                0,
                color,
                background_color,
                bold,
                italic,
                underline,
                strikethrough,
            )
        else:
            sty = self._style
        sty_msg = stylize(msg, sty)
        padded = self.padding(sty_msg, sty)
        if space:
            self._ensure_linespaces()
        self._update_lastchar(sty_msg)
        self._console.print(padded)

    def print_renderable(
        self, rd: RenderableType, *, space: bool = True, indentation: int = 0
    ) -> None:
        msg = self.padding(rd, TextStyle(indentation=indentation))
        if space:
            self._ensure_linespaces()
        self._update_lastchar(msg)
        self._console.print(msg)

    def err(self, msg: str) -> None:
        self.print(f"✖  {msg}", color="red1", bold=True)

    def fatal(self, msg: str) -> None:
        self.err(msg)
        exit(1)

    def die(self, msg: str | Exception) -> None:
        if isinstance(msg, Exception):
            from rich.traceback import install

            install()
            self._console.print_exception()
            exit(1)

        self.fatal(msg)

    def success(self, msg: str) -> None:
        self.print(f"✔  {msg}", color="green", bold=True)

    def prompt(
        self,
        msg: str,
        choices: list[str] | None = None,
        *,
        space: bool = True,
        **kwargs,
    ) -> str:
        if kwargs:
            sty = TextStyle(**kwargs)
        else:
            sty = self._style
        sty_msg = stylize(msg, sty)
        if space:
            self._ensure_linespaces()
        self._update_lastchar(sty_msg)
        return self._prompt.ask(sty_msg, console=self._console, choices=choices)

    def _update_lastchar(self, msg) -> None:
        try:
            self._lastchar = str(msg)[-1]
        except IndexError:
            self._lastchar = ""

    def _ensure_linespaces(self) -> None:
        if not self._lastchar in {"\n", "", " "}:
            for _ in range(self._linespaces):
                self._console.print("")


_default_tty = RichTty()
debug = _default_tty.debug
info = _default_tty.info
warning = _default_tty.warning
error = _default_tty.error
err = _default_tty.err
critical = _default_tty.critical
print = _default_tty.print
print_renderable = _default_tty.print_renderable
success = _default_tty.success
die = _default_tty.die
fatal = _default_tty.fatal
prompt = _default_tty.prompt


def set_tty(tty: RichTty) -> None:
    global _default_tty
    _default_tty = tty


def configure_logger(
    *,
    console_out: bool = True,
    level: int = logging.WARNING,
    log_format: str = "%(asctime)s - %(levelname)s - %(message)s",
    datefmt="[%Y-%m-%d %H:%M:%S]",
    filename: Optional[str] = None,
    console_handler: Optional[logging.Handler] = None,
) -> None:
    handlers: list[logging.Handler] = []

    if console_out and not console_handler:
        console_handler = logging.StreamHandler()

    if filename:
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        handlers.append(file_handler)

    if console_handler:
        console_handler.setLevel(level)
        handlers.append(console_handler)

    logging.basicConfig(
        level=level,
        format=log_format,
        datefmt=datefmt,
        handlers=handlers,
    )
