#
# Author: Francesco Talpo
#
# This file is part of Modplanner.
#
# Modplanner is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Modplanner is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Modplanner. If not, see <https://www.gnu.org/licenses/>.
#


"""
Module containing some utility functions.
"""


from __future__ import annotations

import json
import operator
import os
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from modplanner.errors import ModplannerError

if TYPE_CHECKING:
    from typing import Any, Callable, Type

    from filetree import PathLike


def load_yaml(path: PathLike) -> dict:
    with open(path, "r") as f:
        return yaml.safe_load(f)


def add_serializer(obj_type: Type[Any], dict_factory: Callable[[Any], dict]) -> None:
    _srl = lambda dumper, obj: dumper.represent_dict(dict_factory(obj))
    yaml.add_representer(obj_type, _srl)


def dumps_yaml(dct: dict) -> str:
    return yaml.dump(dct)


class BadModuleName(ModplannerError):
    """
    Badly formatted module name.
    """

    def __init__(self, name: str, *args) -> None:
        self.name = name
        super().__init__(*args)


def validate_module_name(name: str) -> None:
    if "/" not in name:
        raise BadModuleName(name)


class SourceError(ModplannerError):
    """
    Error sourcing file.
    """


def source(sourcefile: PathLike) -> None:
    sourcefile = Path(sourcefile)
    if not sourcefile.is_file():
        raise FileNotFoundError(f"Missing file to source: {sourcefile}")

    src = "." + f" {sourcefile}" + " > /dev/null"
    dump = (
        sys.executable + r" -c \"import os, json; print(json.dumps(dict(os.environ)))\""
    )
    try:
        pipe = subprocess.run(
            f'/bin/bash -c "{src} && {dump}"',
            capture_output=True,
            text=True,
            shell=True,
        )
        out = pipe.stdout
        env = json.loads(out.strip())
        os.environ.update(env)
    except json.JSONDecodeError:
        raise SourceError(f"Error sourcing file '{sourcefile}': invalid JSON output")


class ModuleLoadError(ModplannerError):
    """
    Error in module load.
    """


def load_module(module: str) -> None:
    load = f"module load {module} > /dev/null"
    dump = (
        sys.executable + r" -c \"import os, json; print(json.dumps(dict(os.environ)))\""
    )
    pipe = subprocess.run(
        f'/bin/bash -c "{load} && {dump}"',
        capture_output=True,
        text=True,
        shell=True,
    )
    out = pipe.stdout
    try:
        env = json.loads(out.strip())
        os.environ.update(env)
    except json.JSONDecodeError:
        raise ModuleLoadError(f"Error loading module '{module}': invalid JSON output")


class Version:

    def __init__(self, version: str) -> None:
        self._parts = []
        for part in version.split("."):
            try:
                self._parts.append(int(part))
            except ValueError:
                raise ValueError(f"Invalid version: {version}.")

    @property
    def parts(self) -> list[int]:
        return self._parts

    @staticmethod
    def _flatten(one: list[int], two: list[int]) -> tuple[list[int], list[int]]:
        ret_one, ret_two = one.copy(), two.copy()
        if len(one) == len(two):
            return ret_one, ret_two

        elif len(one) < len(two):
            diff = len(two) - len(one)
            for _ in range(diff):
                ret_one.append(0)
        elif len(one) > len(two):
            diff = len(one) - len(two)
            for _ in range(diff):
                ret_two.append(0)

        return ret_one, ret_two

    def _compare(self, other, op) -> bool:
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            raise ValueError("Bad comparison.")

        self_flat, other_flat = self._flatten(self.parts, other.parts)
        for i, j in zip(self_flat, other_flat):
            if op(i, j):
                return True
        return False

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False

        self_flat, other_flat = self._flatten(self.parts, other.parts)
        for i, j in zip(self_flat, other_flat):
            if i != j:
                return False
        return True

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other) -> bool:
        return self._compare(other, operator.lt)

    def __gt__(self, other) -> bool:
        return self._compare(other, operator.gt)

    def __le__(self, other) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other) -> bool:
        return self.__gt__(other) or self.__eq__(other)
