from pathlib import Path

from modplanner.service.map.context import (
    ModuleContextBuilder,
    ModuleHideContext,
    ModuleMapContext,
    MappingRule,
)
from modplanner.module import Module


import pytest


@pytest.fixture
def module():
    return Module("pack1", Path("path1"))


class TestMapContext:

    def test_module_is_not_mapped(self, module):
        ctx = ModuleMapContext()
        assert not ctx.is_mapped(module, {})

    def test_module_add_mapping(self, module):
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        assert module in ctx._mappings and ctx._mappings[module] == {
            "mapping": {"value"}
        }

    def test_module_is_mapped_empty_returns_false(self, module):
        ctx = ModuleMapContext()
        assert not ctx.is_mapped(module, {})

    def test_module_is_mapped_true(self, module):
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        assert ctx.is_mapped(module, {"mapping": "value"})

    def test_module_is_mapped_false(self, module):
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        assert not ctx.is_mapped(module, {"not": "mapped"})

    def test_has_mapping_true(self, module):
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        assert ctx.has_mapping(module, "mapping")

    def test_has_mapping_false(self, module):
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        assert not ctx.has_mapping(module, "nope")

    def test_get_mappings(self, module):
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        assert ctx.get_mappings(module) == {"mapping": {"value"}}

    def test_get_mapped_modules(self, module):
        other_module = Module("pack2", Path("path2"))
        ctx = ModuleMapContext()
        ctx.map_module(module, "mapping", "value")
        ctx.map_module(other_module, "other_mapping", "other_value")
        assert list(ctx.mapped_modules({"mapping": "value"})) == [module]


class TestHideContext:

    def test_hide_is_hidden_module(self, module):
        ctx = ModuleHideContext()
        ctx.hide(module)
        assert ctx.is_hidden(module)

    def test_hide_show(self, module):
        ctx = ModuleHideContext()
        ctx.hide(module)
        assert ctx.is_hidden(module)
        ctx.show(module)
        assert not ctx.is_hidden(module)

    def test_hide_condition_not_globally_hidden(self, module):
        ctx = ModuleHideContext()
        ctx.add_condition(module, lambda ctx: ctx["mapping"] == "value")
        assert not ctx.is_hidden(module)

    def test_hide_condition_specific_context(self, module):
        ctx = ModuleHideContext()
        ctx.add_condition(module, lambda ctx: ctx["mapping"] == "value")
        assert ctx.is_hidden(module, context={"mapping": "value"})

    def test_show_does_not_clear_conditions(self, module):
        ctx = ModuleHideContext()
        ctx.add_condition(module, lambda ctx: ctx["mapping"] == "value")
        assert ctx.is_hidden(module, context={"mapping": "value"})
        ctx.show(module)
        assert ctx.is_hidden(module, context={"mapping": "value"})

    def test_show_does_clears_conditions(self, module):
        ctx = ModuleHideContext()
        ctx.add_condition(module, lambda ctx: ctx["mapping"] == "value")
        assert ctx.is_hidden(module, context={"mapping": "value"})
        ctx.show(module, clear_conditions=True)
        assert not ctx.is_hidden(module, context={"mapping": "value"})


class FakeFilter:

    def __init__(self, names):
        self.seen = set()
        self.names = names

    def __call__(self, module: Module) -> bool:
        if module.package_name in self.names:
            self.seen.add(module.package_name)
            return True
        return False


class FakeCallback:

    def __init__(self):
        self.seen = set()

    def __call__(self, module: Module) -> None:
        self.seen.add(module.package_name)


@pytest.fixture
def modules():
    return [
        Module("pack1", Path("path1")),
        Module("pack2", Path("path2")),
        Module("pack3", Path("path3")),
    ]


def rule(name, exclusive=False, deferred=False):
    flt = FakeFilter([name])
    cb = FakeCallback()
    r = MappingRule([flt], [cb], exclusive=exclusive, deferred=deferred)
    return flt, cb, r


class TestContextBuilder:

    # FIXME: this code is beyond ugly, and I think that it's an indication
    # that the ContextBuilder code is too rigid; still, I can't find a better structure.

    def test_exclusive_rule_removes_from_pool(self, modules):
        flt1, _, r1 = rule("pack1", exclusive=True)
        flt2, _, r2 = rule("pack1")
        ctx_builder = ModuleContextBuilder()
        ctx_builder.add_global_rule(r1)
        ctx_builder.add_global_rule(r2)
        _ = ctx_builder.build_context(modules)
        assert "pack1" in flt1.seen and not "pack1" in flt2.seen

    def test_non_exclusive_rule_does_not_remove_from_pool(self, modules):
        flt1, _, r1 = rule("pack1")
        flt2, _, r2 = rule("pack1")
        ctx_builder = ModuleContextBuilder()
        ctx_builder.add_global_rule(r1)
        ctx_builder.add_global_rule(r2)
        _ = ctx_builder.build_context(modules)
        assert "pack1" in flt1.seen and "pack1" in flt2.seen

    def test_global_rules_are_first(self, modules):
        flt1, _, r1 = rule("pack1", exclusive=True)
        flt2, _, r2 = rule("pack1")
        ctx_builder = ModuleContextBuilder()
        ctx_builder.add_global_rule(r1)
        ctx_builder.add_local_rules([r2])
        _ = ctx_builder.build_context(modules)
        assert "pack1" in flt1.seen and "pack1" not in flt2.seen

    def test_local_pool_exclusive_rules_remove_from_local_pool(self, modules):
        flt1, _, r1 = rule("pack1", exclusive=True)
        flt2, _, r2 = rule("pack1")
        ctx_builder = ModuleContextBuilder()
        ctx_builder.add_local_rules([r1, r2])
        _ = ctx_builder.build_context(modules)
        assert "pack1" in flt1.seen and "pack1" not in flt2.seen

    def test_local_pool_is_reset_for_each_batch(self, modules):
        flt1, _, r1 = rule("pack1", exclusive=True)
        flt2, _, r2 = rule("pack1", exclusive=True)
        ctx_builder = ModuleContextBuilder()
        ctx_builder.add_local_rules([r1])
        ctx_builder.add_local_rules([r2])
        _ = ctx_builder.build_context(modules)
        assert "pack1" in flt1.seen and "pack1" in flt2.seen
